#define _WIN32_WINNT 0x0501
#define _CRT_SECURE_NO_WARNINGS

#define WIN32_LEAN_AND_MEAN
#define WIN32_EXTRA_LEAN
#include <windows.h>

#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>

#define BIG_PATH	32767
#define TEMP_SIZE	1024
#define USER_BUF_SZ	128

// Runs a system command, substituting in a file name for %s
int sys(const TCHAR *cmd, const TCHAR *file) {
	TCHAR temp[BIG_PATH+50];
	int retval;
	_sntprintf(temp, BIG_PATH+50, cmd, file);
	retval = _tsystem(temp);
	if (retval != 0) {
		_ftprintf(stderr, TEXT("* Command failed (%d): %s\n"), retval, temp);
	}
	return retval;
}

#include "mingw-unicode.c"
int _tmain(int argc, _TCHAR* argv[]) {
	wchar_t *src, *dest, temp[TEMP_SIZE], name[USER_BUF_SZ], domain[USER_BUF_SZ];
	int retval = 0;
	DWORD size = 0, size2 = USER_BUF_SZ;
	SECURITY_DESCRIPTOR *owner, *dacl;
	PSID sid;
	BOOL def;
	SID_NAME_USE use;

	if (argc != 3) {
		_ftprintf(stderr, TEXT("You need to provide a source and destination file\n"));
		return 1;
	}

	src = argv[1];
	dest = argv[2];

	// Get the current owner of the destination file
	GetFileSecurity(dest, OWNER_SECURITY_INFORMATION, NULL, 0, &size);
	owner = (SECURITY_DESCRIPTOR*)malloc(size);
	if (!GetFileSecurity(dest, OWNER_SECURITY_INFORMATION, owner, size, &size)) {
		_ftprintf(stderr, TEXT("Failed to get current owner of '%s': %d\n"), dest, GetLastError());
		return 1;
	}
	if (!IsValidSecurityDescriptor(owner)) {
		_ftprintf(stderr, TEXT("Owner of '%s' is not valid\n"), dest);
		return 1;
	}
	if (!GetSecurityDescriptorOwner(owner, &sid, &def)) {
		_ftprintf(stderr, TEXT("GetSecurityDescriptorOwner failed: %d\n"), GetLastError());
		return 1;
	}
	size = USER_BUF_SZ;
	if (!LookupAccountSid(NULL, sid, name, &size, domain, &size2, &use)) {
		_ftprintf(stderr, TEXT("LookupAccountSid failed: %d\n"), GetLastError());
		return 1;
	}

	// Get the current DACLs of the destination file
	GetFileSecurity(dest, DACL_SECURITY_INFORMATION, NULL, 0, &size);
	dacl = (SECURITY_DESCRIPTOR*)malloc(size);
	if (!GetFileSecurity(dest, DACL_SECURITY_INFORMATION, dacl, size, &size)) {
		_ftprintf(stderr, TEXT("Failed to get current DACL of '%s': %d\n"), dest, GetLastError());
		return 1;
	}

	// Take ownership, get rights
	sys(TEXT("takeown /F \"%s\" >nul"), dest);
	sys(TEXT("icacls \"%s\" /grant \"%%USERNAME%%\":F >nul"), dest);

	// Replace the file
	if (CopyFile(src, dest, FALSE)) {
		_tprintf(TEXT("Copied '%s' to '%s'\n"), src, dest);
	} else {
		_ftprintf(stderr, TEXT("Error coping file '%s' to '%s': %d\n"), src, dest, GetLastError());
		retval = 1;
	}

	// Set the owner back to the original owner
	_sntprintf(temp, TEMP_SIZE, TEXT("icacls \"%%s\" /setowner \"%s\\%s\" >nul"), domain, name);
	sys(temp, dest);

	// Set the DACLs back to how they were
	if (!SetFileSecurity(dest, DACL_SECURITY_INFORMATION, dacl)) {
		_ftprintf(stderr, TEXT("Failed to set current DACL of '%s': %d\n"), dest, GetLastError());
	}

	free(owner);
	free(dacl);

	return retval;
}
